import { date } from 'quasar';

/** An new event does not contain an id and the type might be unset */
export type EditableEvent = Omit<Omit<Omit<FG.Event, 'jobs'>, 'type'>, 'id'> & {
  type?: FG.EventType | number;
  id?: number;
  jobs: Job[];
};

export class Job implements FG.Job {
  id = NaN;
  start: Date;
  end?: Date = undefined;
  type: FG.JobType | number = NaN;
  comment?: string = undefined;
  locked = false;
  services = [] as FG.Service[];
  required_services = 0;

  /**
   * Build Job from API Job interface
   * @param iJob Object following the API Job interface
   */
  constructor(iJob?: Partial<FG.Job>) {
    if (!iJob || iJob.start === undefined)
      this.start = date.buildDate({
        hours: new Date().getHours(),
        minutes: 0,
        seconds: 0,
        milliseconds: 0,
      });
    else this.start = new Date(); // <-- make TS happy "no initalizer"
    if (!iJob || iJob.end === undefined) {
      this.end = date.buildDate({
        hours: new Date().getHours() + 4,
        minutes: 0,
        seconds: 0,
        milliseconds: 0,
      });
    }
    if (iJob !== undefined) Object.assign(this, iJob);
  }

  /**
   * Create Job from start Date
   * @param start when does the event start?
   * @param adjustTime Set hours to current value, zero minutes and seconds
   * @param duration How long should the job go? Value in hours or undefined
   * @returns new Job event
   */
  static fromDate(start: Date, adjustTime = true, duration?: number) {
    if (adjustTime)
      start = date.adjustDate(start, {
        hours: new Date().getHours(),
        minutes: 0,
        seconds: 0,
        milliseconds: 0,
      });
    return new Job({
      start: start,
      end: duration === undefined ? undefined : date.addToDate(start, { hours: duration }),
    });
  }

  /**
   * Check if this instance was loaded from API
   */
  isPersistent() {
    return !isNaN(this.id);
  }
}

export function emptyEvent(startDate: Date = new Date()): EditableEvent {
  return {
    start: date.adjustDate(startDate, { hours: 0, minutes: 0, seconds: 0, milliseconds: 0 }),
    jobs: [Job.fromDate(startDate, true, 4)],
    is_template: false,
  };
}
